package Linode::Longview::DataGetter::Disk;

use Exporter;
@ISA       = qw(Exporter);
@EXPORT_OK = qw(get);

require "syscall.ph";
use Linode::Longview::Util qw(slurp_file detect_system $procfs $sysfs $arch);
use POSIX qw(uname);
use Cwd qw(abs_path);

sub get {
	my $dataref = shift;
	my @swaps   = @{ _get_swap_info() };
	_get_mounted_info($dataref);

	# OpenVZ doesn't do /proc/diskstats, bail
	return $dataref if ( detect_system() eq 'openvz' );

	# get the information from /proc/diskstats
	@data = slurp_file( $procfs . 'diskstats' )
		or die "Couldn't read ${procfs}diststats: $!";
	foreach my $line (@data) {
		$line =~ s/^\s*//;

		#  202 0 xvda 3125353 13998 4980 2974 366 591 760 87320 15 366 9029
		my ( $device, $reads, $writes ) = ( split( /\s+/, $line ) )[ 2, 3, 7 ];
		$device = '/dev/' . $device;
		next unless ( $reads || $writes );

		$dataref->{LONGTERM}->{"Disk.$device.reads"}  = $reads + 0;
		$dataref->{LONGTERM}->{"Disk.$device.writes"} = $writes + 0;
		$dataref->{INSTANT}->{"Disk.$device.isswap"} = $device ~~ @swaps ? 1 : 0;
	}

	return $dataref;
}

sub _get_swap_info {
	my @proc_swaps = slurp_file( $procfs . 'swaps' )
		or die "Couldn't read ${procfs}swaps: $!";

	# first line of /proc/swaps is a header
	shift @proc_swaps;

	my @swaps;
	foreach my $line (@proc_swaps) {

		# /dev/xvdb partition 262140 122988 -1
		( $device, $type, $size, $used ) = ( split( /\s+/, $line ) )[ 0 .. 3 ];

		# not tracking file-backed swaps
		next if $type ne "partition";
		push @swaps, $device;
	}
	return \@swaps;
}

sub _get_mounted_info {
	my $dataref = shift;
	my %devices;
	my @mtab = slurp_file('/etc/mtab') or die "Couldn't read /etc/mtab: $!";
	foreach my $line ( grep {m|^/|} @mtab ) {
		my ( $device, $path ) = ( $line =~ m|^(\S+)\s+(\S+)| );
		next unless $device && $path;

		if ( $device =~ m|^/dev/mapper| ) {
			$device = abs_path( '/dev/mapper/' . readlink($device) );
		}

		if ($device eq '/dev/root') {
			my $root_dev = slurp_file($procfs . 'cmdline');
			($device) = ($root_dev =~ m|root=(\S+)|);
		}

		my ( $type, $bsize, $blocks, $bfree, $bavail, $files, $ffree )
			= _statfs($path);

		$dataref->{LONGTERM}->{"Disk.$device.fs.free"}   = $bfree * $bsize;
		$dataref->{LONGTERM}->{"Disk.$device.fs.total"}  = $bsize * $blocks;
		$dataref->{LONGTERM}->{"Disk.$device.fs.ifree"}  = $ffree;
		$dataref->{LONGTERM}->{"Disk.$device.fs.itotal"} = $files;
		$dataref->{INSTANT}->{"Disk.$device.fs.path"}    = $path;
	}
	return $dataref;
}

sub _statfs {
	my $path = shift or die "Nothing to stat: $!";
	my $fmt = "\0" x 512;
	syscall( &SYS_statfs, $path, $fmt ) == 0
		or die "Couldn't call statfs syscall: $!";

	# unsigned long vs unsigned quad on 32 vs 64 bit boxes
	if ( $arch eq 'i686' ) {
		return unpack "L7", $fmt;
	}
	elsif ( $arch eq 'x86_64' ) {
		return unpack "Q7", $fmt;
	}
	else {
		die "Unable to determine architecture";
	}
}

1;
